# Generated by Django 3.2.15 on 2022-08-05 15:13

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Lock',
            fields=[
                ('token', models.CharField(max_length=128, primary_key=True, serialize=False)),
                ('data', models.JSONField()),
            ],
            options={
                'db_table': 'manabi_lock',
            },
        ),
    ]
